"""
Author: Michel Peltriaux
Organization: Struktur- und Genehmigungsdirektion Nord, Rhineland-Palatinate, Germany
Contact: ksp-servicestelle@sgdnord.rlp.de
Created on: 19.08.22

"""
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Sum
from django.http import HttpRequest, JsonResponse
from django.shortcuts import get_object_or_404, render, redirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from compensation.forms.compensation import EditCompensationForm, NewCompensationForm
from compensation.models import Compensation
from compensation.tables.compensation import CompensationTable
from intervention.models import Intervention
from konova.contexts import BaseContext
from konova.decorators import shared_access_required, default_group_required, any_group_check, login_required_modal, \
    uuid_required
from konova.forms import SimpleGeomForm
from konova.forms.modals import RemoveModalForm
from konova.settings import DEFAULT_GROUP, ZB_GROUP, ETS_GROUP
from konova.sub_settings.context_settings import TAB_TITLE_IDENTIFIER
from konova.utils.message_templates import COMPENSATION_REMOVED_TEMPLATE, DATA_CHECKED_PREVIOUSLY_TEMPLATE, \
    RECORDED_BLOCKS_EDIT, CHECK_STATE_RESET, FORM_INVALID, PARAMS_INVALID, IDENTIFIER_REPLACED, \
    COMPENSATION_ADDED_TEMPLATE, DO_NOT_FORGET_TO_SHARE, GEOMETRY_SIMPLIFIED


@login_required
@any_group_check
def index_view(request: HttpRequest):
    """
    Renders the index view for compensation

    Args:
        request (HttpRequest): The incoming request

    Returns:
        A rendered view
    """
    template = "generic_index.html"
    compensations = Compensation.objects.filter(
        deleted=None,  # only show those which are not deleted individually
        intervention__deleted=None,  # and don't show the ones whose intervention has been deleted
    ).order_by(
        "-modified__timestamp"
    )
    table = CompensationTable(
        request=request,
        queryset=compensations
    )
    context = {
        "table": table,
        TAB_TITLE_IDENTIFIER: _("Compensations - Overview"),
    }
    context = BaseContext(request, context).context
    return render(request, template, context)


@login_required
@default_group_required
@shared_access_required(Intervention, "intervention_id")
def new_view(request: HttpRequest, intervention_id: str = None):
    """
    Renders a view for a new compensation creation

    Args:
        request (HttpRequest): The incoming request

    Returns:

    """
    template = "compensation/form/view.html"
    if intervention_id is not None:
        try:
            intervention = Intervention.objects.get(id=intervention_id)
        except ObjectDoesNotExist:
            messages.error(request, PARAMS_INVALID)
            return redirect("home")
        if intervention.is_recorded:
            messages.info(
                request,
                RECORDED_BLOCKS_EDIT
            )
            return redirect("intervention:detail", id=intervention_id)

    data_form = NewCompensationForm(request.POST or None, intervention_id=intervention_id)
    geom_form = SimpleGeomForm(request.POST or None, read_only=False)
    if request.method == "POST":
        if data_form.is_valid() and geom_form.is_valid():
            generated_identifier = data_form.cleaned_data.get("identifier", None)
            comp = data_form.save(request.user, geom_form)
            if generated_identifier != comp.identifier:
                messages.info(
                    request,
                    IDENTIFIER_REPLACED.format(
                        generated_identifier,
                        comp.identifier
                    )
                )
            messages.success(request, COMPENSATION_ADDED_TEMPLATE.format(comp.identifier))
            if geom_form.geometry_simplified:
                messages.info(
                    request,
                    GEOMETRY_SIMPLIFIED
                )
            return redirect("compensation:detail", id=comp.id)
        else:
            messages.error(request, FORM_INVALID, extra_tags="danger",)
    else:
        # For clarification: nothing in this case
        pass
    context = {
        "form": data_form,
        "geom_form": geom_form,
        TAB_TITLE_IDENTIFIER: _("New compensation"),
    }
    context = BaseContext(request, context).context
    return render(request, template, context)


@login_required
@default_group_required
def new_id_view(request: HttpRequest):
    """ JSON endpoint

    Provides fetching of free identifiers for e.g. AJAX calls

    """
    tmp = Compensation()
    identifier = tmp.generate_new_identifier()
    while Compensation.objects.filter(identifier=identifier).exists():
        identifier = tmp.generate_new_identifier()
    return JsonResponse(
        data={
            "gen_data": identifier
        }
    )


@login_required
@default_group_required
@shared_access_required(Compensation, "id")
def edit_view(request: HttpRequest, id: str):
    """
    Renders a view for editing compensations

    Args:
        request (HttpRequest): The incoming request

    Returns:

    """
    template = "compensation/form/view.html"
    # Get object from db
    comp = get_object_or_404(Compensation, id=id)
    if comp.is_recorded:
        messages.info(
            request,
            RECORDED_BLOCKS_EDIT
        )
        return redirect("compensation:detail", id=id)

    # Create forms, initialize with values from db/from POST request
    data_form = EditCompensationForm(request.POST or None, instance=comp)
    geom_form = SimpleGeomForm(request.POST or None, read_only=False, instance=comp)
    if request.method == "POST":
        if data_form.is_valid() and geom_form.is_valid():
            # Preserve state of intervention checked to determine whether the user must be informed or not
            # about a change of the check state
            intervention_is_checked = comp.intervention.checked is not None

            # The data form takes the geom form for processing, as well as the performing user
            comp = data_form.save(request.user, geom_form)
            if intervention_is_checked:
                messages.info(request, CHECK_STATE_RESET)
            messages.success(request, _("Compensation {} edited").format(comp.identifier))
            if geom_form.geometry_simplified:
                messages.info(
                    request,
                    GEOMETRY_SIMPLIFIED
                )
            return redirect("compensation:detail", id=comp.id)
        else:
            messages.error(request, FORM_INVALID, extra_tags="danger",)
    else:
        # For clarification: nothing in this case
        pass
    context = {
        "form": data_form,
        "geom_form": geom_form,
        TAB_TITLE_IDENTIFIER: _("Edit {}").format(comp.identifier),
    }
    context = BaseContext(request, context).context
    return render(request, template, context)


@login_required
@any_group_check
@uuid_required
def detail_view(request: HttpRequest, id: str):
    """ Renders a detail view for a compensation

    Args:
        request (HttpRequest): The incoming request
        id (str): The compensation's id

    Returns:

    """
    template = "compensation/detail/compensation/view.html"
    comp = get_object_or_404(
        Compensation.objects.select_related(
            "modified",
            "created",
            "geometry"
        ),
        id=id,
        deleted=None,
        intervention__deleted=None,
    )
    geom_form = SimpleGeomForm(instance=comp)
    parcels = comp.get_underlying_parcels()
    _user = request.user
    is_data_shared = comp.intervention.is_shared_with(_user)

    # Order states according to surface
    before_states = comp.before_states.all().prefetch_related("biotope_type").order_by("-surface")
    after_states = comp.after_states.all().prefetch_related("biotope_type").order_by("-surface")
    actions = comp.actions.all().prefetch_related("action_type")

    # Precalculate logical errors between before- and after-states
    # Sum() returns None in case of no states, so we catch that and replace it with 0 for easier handling
    sum_before_states = comp.get_surface_before_states()
    sum_after_states = comp.get_surface_after_states()
    diff_states = abs(sum_before_states - sum_after_states)

    request = comp.set_status_messages(request)

    last_checked = comp.intervention.get_last_checked_action()
    last_checked_tooltip = ""
    if last_checked:
        last_checked_tooltip = DATA_CHECKED_PREVIOUSLY_TEMPLATE.format(last_checked.get_timestamp_str_formatted(), last_checked.user)

    requesting_user_is_only_shared_user = comp.is_only_shared_with(_user)
    if requesting_user_is_only_shared_user:
        messages.info(
            request,
            DO_NOT_FORGET_TO_SHARE
        )

    context = {
        "obj": comp,
        "last_checked": last_checked,
        "last_checked_tooltip": last_checked_tooltip,
        "geom_form": geom_form,
        "parcels": parcels,
        "is_entry_shared": is_data_shared,
        "actions": actions,
        "before_states": before_states,
        "after_states": after_states,
        "sum_before_states": sum_before_states,
        "sum_after_states": sum_after_states,
        "diff_states": diff_states,
        "is_default_member": _user.in_group(DEFAULT_GROUP),
        "is_zb_member": _user.in_group(ZB_GROUP),
        "is_ets_member": _user.in_group(ETS_GROUP),
        "LANIS_LINK": comp.get_LANIS_link(),
        TAB_TITLE_IDENTIFIER: f"{comp.identifier} - {comp.title}",
        "has_finished_deadlines": comp.get_finished_deadlines().exists(),
    }
    context = BaseContext(request, context).context
    return render(request, template, context)


@login_required_modal
@login_required
@default_group_required
@shared_access_required(Compensation, "id")
def remove_view(request: HttpRequest, id: str):
    """ Renders a modal view for removing the compensation

    Args:
        request (HttpRequest): The incoming request
        id (str): The compensation's id

    Returns:

    """
    comp = get_object_or_404(Compensation, id=id)
    form = RemoveModalForm(request.POST or None, instance=comp, request=request)
    return form.process_request(
        request=request,
        msg_success=COMPENSATION_REMOVED_TEMPLATE.format(comp.identifier),
        redirect_url=reverse("compensation:index"),
    )

