"""
Author: Michel Peltriaux
Organization: Struktur- und Genehmigungsdirektion Nord, Rhineland-Palatinate, Germany
Contact: ksp-servicestelle@sgdnord.rlp.de
Created on: 19.08.22

"""
from django.http import HttpRequest
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from ema.models import Ema
from konova.contexts import BaseContext
from konova.decorators import uuid_required
from konova.forms import SimpleGeomForm
from konova.sub_settings.context_settings import TAB_TITLE_IDENTIFIER
from konova.utils.generators import generate_qr_code

@uuid_required
def report_view(request:HttpRequest, id: str):
    """ Renders the public report view

    Args:
        request (HttpRequest): The incoming request
        id (str): The id of the intervention

    Returns:

    """
    # Reuse the compensation report template since EMAs are structurally identical
    template = "ema/report/report.html"
    ema = get_object_or_404(Ema, id=id)

    tab_title = _("Report {}").format(ema.identifier)
    # If intervention is not recorded (yet or currently) we need to render another template without any data
    if not ema.is_ready_for_publish():
        template = "report/unavailable.html"
        context = {
            TAB_TITLE_IDENTIFIER: tab_title,
        }
        context = BaseContext(request, context).context
        return render(request, template, context)

    # Prepare data for map viewer
    geom_form = SimpleGeomForm(
        instance=ema,
    )
    parcels = ema.get_underlying_parcels()

    qrcode_url = request.build_absolute_uri(reverse("ema:report", args=(id,)))
    qrcode_img = generate_qr_code(qrcode_url, 10)
    qrcode_lanis_url = ema.get_LANIS_link()
    qrcode_img_lanis = generate_qr_code(qrcode_lanis_url, 7)

    # Order states by surface
    before_states = ema.before_states.all().order_by("-surface").prefetch_related("biotope_type")
    after_states = ema.after_states.all().order_by("-surface").prefetch_related("biotope_type")
    actions = ema.actions.all().prefetch_related("action_type")

    context = {
        "obj": ema,
        "qrcode": {
            "img": qrcode_img,
            "url": qrcode_url
        },
        "qrcode_lanis": {
            "img": qrcode_img_lanis,
            "url": qrcode_lanis_url
        },
        "is_entry_shared": False,  # disables action buttons during rendering
        "before_states": before_states,
        "after_states": after_states,
        "geom_form": geom_form,
        "parcels": parcels,
        "actions": actions,
        "tables_scrollable": False,
        TAB_TITLE_IDENTIFIER: tab_title,
    }
    context = BaseContext(request, context).context
    return render(request, template, context)
